import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { Joint } from '../../lab/engine2D/Joint.js';
import { JointUtil } from '../../lab/engine2D/JointUtil.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Scrim } from '../../lab/engine2D/Scrim.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
import { Walls } from '../../lab/engine2D/Walls.js';
export class CartPendulum2App extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-3, -2, 3, 2);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.stiffness = 6;
        this.springDamping = 0;
        this.spring = null;
        this.sim.setShowForces(false);
        this.dampingLaw = new DampingLaw(0, 0.15, this.simList);
        this.gravityLaw = new GravityLaw(9.8, this.simList);
        this.elasticity.setElasticity(0.8);
        this.addPlaybackControls();
        let pn;
        pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, CartPendulum2App.en.SPRING_DAMPING, CartPendulum2App.i18n.SPRING_DAMPING, () => this.getSpringDamping(), a => this.setSpringDamping(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, CartPendulum2App.en.STIFFNESS, CartPendulum2App.i18n.STIFFNESS, () => this.getStiffness(), a => this.setStiffness(a)));
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
        this.configure();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'CartPendulum2App';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw', myName + '.');
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    configure() {
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.advance.reset();
        const wallPivotX = -0.85;
        const wallPivotY = 0;
        const displacement = 0.13;
        const cart = Shapes.makeBlock2(0.3, 1.0, CartPendulum2App.en.CART, CartPendulum2App.i18n.CART);
        const pivotX = 0.5 * cart.getWidth();
        const pivot1Y = 0.85 * cart.getHeight();
        const pivot2Y = 0.15 * cart.getHeight();
        const bodyX = 0.3 * cart.getWidth();
        const bodyY = 0.5 * cart.getHeight();
        this.sim.addBody(cart);
        this.displayList.findShape(cart).setFillStyle('rgb(200,200,200)');
        const pendulum = Shapes.makePendulum(0.03, 1.0, 0.15, CartPendulum2App.en.PENDULUM, CartPendulum2App.i18n.PENDULUM);
        this.sim.addBody(pendulum);
        let ds = this.displayList.findShape(pendulum);
        ds.setFillStyle('#B0C4DE');
        ds.setDrawCenterOfMass(true);
        this.spring = new Spring('spring1', cart, new Vector(0.5 * cart.getWidth(), 0.5 * cart.getHeight()), Scrim.getScrim(), new Vector(3, 0), 3.0, this.stiffness);
        this.spring.setDamping(this.springDamping);
        this.sim.addForceLaw(this.spring);
        this.sim.getSimList().add(this.spring);
        let dspr = this.displayList.findSpring(this.spring);
        dspr.setWidth(0.3);
        cart.setPosition(new Vector(1, 0), Math.PI / 2);
        const pivot1_body = new Vector(pivotX, pivot1Y);
        const fixed1_world = cart.bodyToWorld(pivot1_body);
        const j1 = new Joint(cart, pivot1_body, Scrim.getScrim(), fixed1_world, 1, Vector.SOUTH);
        this.sim.addConnector(j1);
        const pivot2_body = new Vector(pivotX, pivot2Y);
        const fixed2_world = cart.bodyToWorld(pivot2_body);
        const j2 = new Joint(cart, pivot2_body, Scrim.getScrim(), fixed2_world, 1, Vector.SOUTH);
        this.sim.addConnector(j2);
        JointUtil.attachRigidBody(this.sim, cart, new Vector(cart.getWidth() / 2, cart.getHeight() / 2), pendulum, new Vector(0, 1), 0);
        this.sim.alignConnectors();
        cart.setZeroEnergyLevel(cart.getPosition().getY());
        pendulum.setZeroEnergyLevel(pendulum.getPosition().getY());
        const Icm = pendulum.momentAboutCM();
        let R = 0.5;
        let r = (Icm / (pendulum.getMass() * R)) + R;
        R = 0.35;
        r = (Icm / (pendulum.getMass() * R)) + R;
        Walls.make(this.sim, 6, 4, 1.0);
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        this.sim.setElasticity(elasticity);
        this.sim.saveInitialState();
        this.clock.setTime(this.sim.getTime());
        this.clock.setRealTime(this.sim.getTime());
        this.easyScript.update();
    }
    ;
    getSpringDamping() {
        return this.springDamping;
    }
    ;
    setSpringDamping(value) {
        if (this.springDamping != value) {
            this.springDamping = value;
            if (this.spring != null) {
                this.spring.setDamping(value);
            }
            this.broadcastParameter(CartPendulum2App.en.SPRING_DAMPING);
        }
    }
    ;
    getStiffness() {
        return this.stiffness;
    }
    ;
    setStiffness(value) {
        if (this.stiffness != value) {
            this.stiffness = value;
            if (this.spring != null) {
                this.spring.setStiffness(value);
            }
            this.broadcastParameter(CartPendulum2App.en.STIFFNESS);
        }
    }
    ;
    graphSetup(_body) {
        const cart = this.sim.getBody('cart');
        const pendulum = this.sim.getBody('pendulum');
        this.graph.line.setXVariable(cart.getVarsIndex() + 0);
        this.graph.line.setYVariable(pendulum.getVarsIndex() + 4);
        this.timeGraph.line1.setYVariable(cart.getVarsIndex() + 0);
        this.timeGraph.line2.setYVariable(pendulum.getVarsIndex() + 4);
    }
    ;
}
CartPendulum2App.en = {
    LENGTH: 'spring length',
    SPRING_DAMPING: 'spring damping',
    STIFFNESS: 'spring stiffness',
    CART: 'cart',
    PENDULUM: 'pendulum'
};
CartPendulum2App.de_strings = {
    LENGTH: 'Federlänge',
    SPRING_DAMPING: 'Federdämpfung',
    STIFFNESS: 'Federsteifheit',
    CART: 'Wagen',
    PENDULUM: 'Pendel'
};
CartPendulum2App.i18n = Util.LOCALE === 'de' ? CartPendulum2App.de_strings : CartPendulum2App.en;
Util.defineGlobal('sims$engine2D$CartPendulum2App', CartPendulum2App);
