import { CheckBoxControl } from '../../lab/controls/CheckBoxControl.js';
import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { ConstantForceLaw } from '../../lab/model/ConstantForceLaw.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { Force } from '../../lab/model/Force.js';
import { GearsConfig } from './GearsConfig.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { JointUtil } from '../../lab/engine2D/JointUtil.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterBoolean, ParameterNumber } from '../../lab/util/Observe.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
import { Walls } from '../../lab/engine2D/Walls.js';
export class GearsApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-4, -6, 8, 6);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.twoGears = true;
        this.turningForce = 0.2;
        this.pinnedGears = true;
        this.gearRight = null;
        this.turnForceLaw = null;
        this.elasticity.setElasticity(0.3);
        this.sim.setShowForces(true);
        this.dampingLaw = new DampingLaw(0, 0.15, this.simList);
        this.gravityLaw = new GravityLaw(0, this.simList);
        this.addPlaybackControls();
        let pb;
        let pn;
        this.addParameter(pb = new ParameterBoolean(this, GearsConfig.en.PINNED_GEARS, GearsConfig.i18n.PINNED_GEARS, () => this.getPinnedGears(), a => this.setPinnedGears(a)));
        this.addControl(new CheckBoxControl(pb));
        this.addParameter(pb = new ParameterBoolean(this, GearsConfig.en.TWO_GEARS, GearsConfig.i18n.TWO_GEARS, () => this.getTwoGears(), a => this.setTwoGears(a)));
        this.addControl(new CheckBoxControl(pb));
        this.addParameter(pn = new ParameterNumber(this, GearsConfig.en.TURNING_FORCE, GearsConfig.i18n.TURNING_FORCE, () => this.getTurningForce(), a => this.setTurningForce(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        this.addControl(new NumericControl(pn));
        pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.makeEasyScript();
        this.addStandardControls();
        this.addURLScriptButton();
        this.config();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'GearsApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw', myName + '.');
        this.terminal.addRegex('GearsApp|Engine2DApp', 'sims$engine2D$', false);
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    graphSetup(body) {
        body = this.sim.getBody(GearsConfig.en.LEFT_GEAR);
        this.graph.line.setXVariable(body.getVarsIndex() + 4);
        this.graph.line.setYVariable(body.getVarsIndex() + 5);
        this.timeGraph.line1.setYVariable(body.getVarsIndex() + 4);
        this.timeGraph.line2.setYVariable(body.getVarsIndex() + 5);
    }
    ;
    config() {
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.advance.reset();
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        this.gearLeft = GearsConfig.makeGear(2.7, [], GearsConfig.en.LEFT_GEAR, GearsConfig.i18n.LEFT_GEAR);
        this.gearLeft.setPosition(new Vector(0, 0), 0);
        this.gearLeft.setMass(1);
        this.sim.addBody(this.gearLeft);
        this.displayList.findShape(this.gearLeft).setFillStyle('lightGray');
        if (this.twoGears) {
            this.gearRight = GearsConfig.makeGear(2.7, [], GearsConfig.en.RIGHT_GEAR, GearsConfig.i18n.RIGHT_GEAR);
            const tooth = 2 * Math.PI / 36;
            this.gearRight.setPosition(new Vector((2 * 2.7) + 0.008 + 0.35, 0), -tooth / 5);
            this.sim.addBody(this.gearRight);
            this.displayList.findShape(this.gearRight).setFillStyle('lightGray');
        }
        else {
            this.gearRight = null;
        }
        if (this.pinnedGears) {
            JointUtil.attachFixedPoint(this.sim, this.gearLeft, Vector.ORIGIN, 1);
            this.gearLeft.setZeroEnergyLevel();
            if (this.gearRight != null) {
                JointUtil.attachFixedPoint(this.sim, this.gearRight, Vector.ORIGIN, 1);
                this.gearRight.setZeroEnergyLevel();
            }
            this.gravityLaw.setGravity(0);
        }
        else {
            const zel = Walls.make(this.sim, 60, 12, 1.0);
            this.gravityLaw.setZeroEnergyLevel(zel);
            this.gravityLaw.setGravity(3);
        }
        this.sim.setElasticity(elasticity);
        this.setTurningForce(this.turningForce);
        this.sim.getVarsList().setTime(0);
        this.sim.saveInitialState();
        this.clock.setTime(0);
        this.clock.setRealTime(0);
        this.easyScript.update();
    }
    ;
    getTurningForce() {
        return this.turningForce;
    }
    ;
    setTurningForce(value) {
        this.turningForce = value;
        if (this.turnForceLaw != null) {
            this.sim.removeForceLaw(this.turnForceLaw);
        }
        if (this.gearLeft) {
            const f = new Force('turning', this.gearLeft, this.gearLeft.getDragPoints()[0], 0, new Vector(value, 0), 0);
            this.turnForceLaw = new ConstantForceLaw(f);
            this.sim.addForceLaw(this.turnForceLaw);
        }
        this.broadcastParameter(GearsConfig.en.TURNING_FORCE);
    }
    ;
    getPinnedGears() {
        return this.pinnedGears;
    }
    ;
    setPinnedGears(value) {
        this.pinnedGears = value;
        this.config();
        this.broadcastParameter(GearsConfig.en.PINNED_GEARS);
    }
    ;
    getTwoGears() {
        return this.twoGears;
    }
    ;
    setTwoGears(value) {
        this.twoGears = value;
        this.config();
        this.broadcastParameter(GearsConfig.en.TWO_GEARS);
    }
    ;
}
Util.defineGlobal('sims$engine2D$GearsApp', GearsApp);
