import { AutoScale } from '../../lab/graph/AutoScale.js';
import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DisplayGraph } from '../../lab/graph/DisplayGraph.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { GraphLine } from '../../lab/graph/GraphLine.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { JointUtil } from '../../lab/engine2D/JointUtil.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { Scrim } from '../../lab/engine2D/Scrim.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
import { Walls } from '../../lab/engine2D/Walls.js';
export class PendulumSpringApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-4, -4, 4, 4);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.springs_ = [];
        this.layout.getSimCanvas().setBackground('black');
        this.sim.setShowForces(false);
        this.dampingLaw = new DampingLaw(0.05, 0.15, this.simList);
        this.sim.addForceLaw(this.dampingLaw);
        this.gravityLaw = new GravityLaw(4, this.simList);
        this.sim.addForceLaw(this.gravityLaw);
        const wallPivotX = 2.0;
        const wallPivotY = 0;
        const map = this.simView.getCoordMap();
        const pendulum = Shapes.makeBlock2(0.3, 1.0, PendulumSpringApp.en.PENDULUM, PendulumSpringApp.i18n.PENDULUM);
        const bodyX = 0.5 * pendulum.getWidth();
        const bodyY = 0.15 * pendulum.getHeight();
        this.sim.addBody(pendulum);
        this.displayList.findShape(pendulum).setFillStyle('#B0C4DE');
        JointUtil.attachRigidBody(this.sim, Scrim.getScrim(), new Vector(wallPivotX, wallPivotY), pendulum, new Vector(bodyX, bodyY), 1);
        const block = Shapes.makeBlock2(0.3, 1.0, PendulumSpringApp.en.BLOCK, PendulumSpringApp.i18n.BLOCK);
        block.setPosition(new Vector(-1.2, -2.5), 0.2);
        this.sim.addBody(block);
        this.displayList.findShape(block).setFillStyle('rgba(255, 165, 0, 0.5)');
        this.rbo.protoSpring.setWidth(0.3);
        this.addSpring(new Spring('spring1', pendulum, new Vector(0.15, 0.7), block, new Vector(0.15, 0.7), 0.75, 1.0));
        this.addSpring(new Spring('spring2', block, new Vector(0.15, 0.7), Scrim.getScrim(), new Vector(-2, 2), 0.75, 1.0));
        this.addSpring(new Spring('spring3', block, new Vector(0.15, 0.3), Scrim.getScrim(), new Vector(-2, -2), 0.75, 1.0));
        const zel = Walls.make2(this.sim, this.simView.getSimRect());
        this.gravityLaw.setZeroEnergyLevel(zel);
        pendulum.setAngle(-Math.PI);
        this.sim.alignConnectors();
        pendulum.setZeroEnergyLevel();
        pendulum.setAngle(0.5);
        this.sim.alignConnectors();
        this.sim.setElasticity(0.3);
        this.sim.saveInitialState();
        this.addPlaybackControls();
        let pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, PendulumSpringApp.en.SPRING_STIFFNESS, PendulumSpringApp.i18n.SPRING_STIFFNESS, () => this.getStiffness(), a => this.setStiffness(a)));
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, PendulumSpringApp.en.SPRING_LENGTH, PendulumSpringApp.i18n.SPRING_LENGTH, () => this.getRestLength(), a => this.setRestLength(a)));
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.graphLine = new GraphLine('TRAILS_LINE', this.sim.getVarsList());
        this.simView.addMemo(this.graphLine);
        this.autoScale = new AutoScale('TRAILS_AUTO_SCALE', this.graphLine, this.simView);
        this.dispGraph = new DisplayGraph(this.graphLine);
        this.dispGraph.setZIndex(3);
        this.dispGraph.setScreenRect(this.simView.getScreenRect());
        this.displayList.add(this.dispGraph);
        this.panZoomParam.setValue(true);
        this.makeEasyScript();
        this.addURLScriptButton();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + ', graphLine: ' + this.graphLine.toStringShort()
            + ', autoScale: ' + this.autoScale.toStringShort()
            + super.toString();
    }
    ;
    addSpring(s) {
        this.springs_.push(s);
        this.sim.addForceLaw(s);
        this.simList.add(s);
    }
    ;
    getClassName() {
        return 'PendulumSpringApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw|graphLine|autoScale', myName + '.');
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.gravityLaw, this.dampingLaw, this.graphLine, this.autoScale);
    }
    ;
    graphSetup(_body) {
        const bodyIdx = this.sim.getBody('block').getVarsIndex();
        this.graphLine.setXVariable(bodyIdx + 0);
        this.graphLine.setYVariable(bodyIdx + 2);
        this.graph.line.setXVariable(bodyIdx + 0);
        this.graph.line.setYVariable(bodyIdx + 2);
        this.timeGraph.line1.setYVariable(bodyIdx + 0);
        this.timeGraph.line2.setYVariable(bodyIdx + 2);
    }
    ;
    getStiffness() {
        return this.springs_[0].getStiffness();
    }
    ;
    setStiffness(value) {
        this.springs_.forEach(s => s.setStiffness(value));
        this.sim.getVarsList().incrSequence(2, 3);
        this.broadcastParameter(PendulumSpringApp.en.SPRING_STIFFNESS);
    }
    ;
    getRestLength() {
        return this.springs_[0].getRestLength();
    }
    ;
    setRestLength(value) {
        this.springs_.forEach(s => s.setRestLength(value));
        this.sim.getVarsList().incrSequence(2, 3);
        this.broadcastParameter(PendulumSpringApp.en.SPRING_LENGTH);
    }
    ;
}
PendulumSpringApp.en = {
    SPRING_LENGTH: 'spring length',
    SPRING_STIFFNESS: 'spring stiffness',
    BLOCK: 'block',
    PENDULUM: 'pendulum'
};
PendulumSpringApp.de_strings = {
    SPRING_LENGTH: 'Federlänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    BLOCK: 'Block',
    PENDULUM: 'Pendel'
};
PendulumSpringApp.i18n = Util.LOCALE === 'de' ? PendulumSpringApp.de_strings : PendulumSpringApp.en;
Util.defineGlobal('sims$engine2D$PendulumSpringApp', PendulumSpringApp);
