import { ButtonControl } from '../../lab/controls/ButtonControl.js';
import { CheckBoxControl } from '../../lab/controls/CheckBoxControl.js';
import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from './Engine2DApp.js';
import { Gravity2Law } from '../../lab/model/Gravity2Law.js';
import { ModifiedEuler } from '../../lab/model/ModifiedEuler.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterBoolean, ParameterNumber } from '../../lab/util/Observe.js';
import { PileConfig } from './PileConfig.js';
import { Polygon } from '../../lab/engine2D/Polygon.js';
import { RandomLCG } from '../../lab/util/Random.js';
import { Util } from '../../lab/util/Util.js';
export class PileAttractApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-3, -3, 3, 3);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.dampingLaw = new DampingLaw(0.05, 0.15);
        this.gravityLaw = new Gravity2Law(1);
        this.numBlocks = 8;
        this.squareBlocks = false;
        this.randomSeed = 0;
        this.buildRNG = new RandomLCG(this.randomSeed);
        this.zeroEnergyLevel = 0;
        this.layout.getSimCanvas().setBackground('black');
        this.rbo.protoDragSpring.setWidth(0.3);
        this.rbo.protoPolygon.setDrawCenterOfMass(true);
        if (0 == 1) {
            this.rbo.protoPolygon.setNameColor('gray');
            this.rbo.protoPolygon.setNameFont('10pt sans-serif');
        }
        this.elasticity.setElasticity(0.8);
        this.sim.setShowForces(false);
        this.sim.setDistanceTol(0.01);
        this.sim.setCollisionAccuracy(0.6);
        this.diffEqSolver.setDiffEqSolver(ModifiedEuler.en.NAME);
        this.addPlaybackControls();
        let pb;
        let pn;
        this.addParameter(pn = new ParameterNumber(this, PileConfig.en.NUM_BLOCKS, PileConfig.i18n.NUM_BLOCKS, () => this.getNumBlocks(), a => this.setNumBlocks(a)));
        pn.setDecimalPlaces(0);
        this.addControl(new NumericControl(pn));
        this.addParameter(pb = new ParameterBoolean(this, PileConfig.en.SQUARE_BLOCKS, PileConfig.i18n.SQUARE_BLOCKS, () => this.getSquareBlocks(), a => this.setSquareBlocks(a)));
        this.addControl(new CheckBoxControl(pb));
        this.addParameter(pn = new ParameterNumber(this, PileConfig.en.RANDOM_SEED, PileConfig.i18n.RANDOM_SEED, () => this.getRandomSeed(), a => this.setRandomSeed(a)));
        pn.setDecimalPlaces(0);
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn = this.gravityLaw.getParameterNumber(Gravity2Law.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        const c = new ButtonControl(PileConfig.i18n.REBUILD, () => this.config());
        this.addControl(c);
        this.makeEasyScript();
        this.addURLScriptButton();
        this.config();
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', dampingLaw: ' + this.dampingLaw.toStringShort()
            + ', gravityLaw: ' + this.gravityLaw.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'PileAttractApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('gravityLaw|dampingLaw', myName + '.');
        this.terminal.addRegex('PileConfig', 'sims$engine2D$', false);
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.dampingLaw, this.gravityLaw);
    }
    ;
    config() {
        this.randomSeed = this.buildRNG.getSeed();
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        Polygon.ID = 1;
        this.advance.reset();
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        const half = Math.floor(this.numBlocks / 2);
        const rest = this.numBlocks - half;
        const blocks = PileConfig.makeRandomBlocks(this.sim, half, -half / 2, 1, this.buildRNG, this.squareBlocks);
        let b2 = PileConfig.makeRandomBlocks(this.sim, rest, -half / 2, -1, this.buildRNG, this.squareBlocks);
        b2.forEach((b) => blocks.push(b));
        blocks.forEach(b => this.displayList.findShape(b).setFillStyle(PileConfig.getRandomColor()));
        this.sim.setElasticity(elasticity);
        this.sim.getVarsList().setTime(0);
        this.sim.saveInitialState();
        this.clock.setTime(0);
        this.clock.setRealTime(0);
        this.easyScript.update();
    }
    ;
    getNumBlocks() {
        return this.numBlocks;
    }
    ;
    setNumBlocks(value) {
        this.numBlocks = value;
        this.config();
        this.broadcastParameter(PileConfig.en.NUM_BLOCKS);
    }
    ;
    getRandomSeed() {
        return this.randomSeed;
    }
    ;
    setRandomSeed(value) {
        this.randomSeed = value;
        this.buildRNG.setSeed(value);
        this.config();
        this.broadcastParameter(PileConfig.en.RANDOM_SEED);
    }
    ;
    getSquareBlocks() {
        return this.squareBlocks;
    }
    ;
    setSquareBlocks(value) {
        this.squareBlocks = value;
        this.config();
        this.broadcastParameter(PileConfig.en.SQUARE_BLOCKS);
    }
    ;
}
Util.defineGlobal('sims$engine2D$PileAttractApp', PileAttractApp);
