import { AbstractApp } from '../common/AbstractApp.js';
import { DisplayLine } from '../../lab/view/DisplayLine.js';
import { DisplayShape } from '../../lab/view/DisplayShape.js';
import { DisplaySpring } from '../../lab/view/DisplaySpring.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { GenericObserver } from '../../lab/util/Observe.js';
import { MoveablePendulumSim } from './MoveablePendulumSim.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { SimList } from '../../lab/model/SimList.js';
import { SimpleAdvance } from '../../lab/model/SimpleAdvance.js';
import { SimRunner } from '../../lab/app/SimRunner.js';
import { SliderControl } from '../../lab/controls/SliderControl.js';
import { Util } from '../../lab/util/Util.js';
export class MoveablePendulumApp extends AbstractApp {
    constructor(elem_ids) {
        Util.setErrorHandler();
        const simRect = new DoubleRect(-3.3, -3.3, 3.3, 3);
        const sim = new MoveablePendulumSim();
        const advance = new SimpleAdvance(sim);
        super(elem_ids, simRect, sim, advance, sim, sim);
        this.simRun.setTimeStep(0.01);
        let ds = new DisplayShape(this.simList.getPointMass('anchor'));
        ds.setFillStyle('');
        ds.setStrokeStyle('red');
        ds.setThickness(4);
        this.anchor = ds;
        this.displayList.add(this.anchor);
        this.rod = new DisplayLine(this.simList.getConcreteLine('rod'));
        this.displayList.add(this.rod);
        this.bob = new DisplayShape(this.simList.getPointMass('bob'));
        this.bob.setFillStyle('blue');
        this.displayList.add(this.bob);
        const dragSpring = sim.getDragSpring();
        const dispSpring = new DisplaySpring(dragSpring);
        dispSpring.setWidth(0.2);
        new GenericObserver(this.simList, (evt) => {
            if (evt.getValue() == dragSpring) {
                if (evt.nameEquals(SimList.OBJECT_ADDED)) {
                    this.displayList.add(dispSpring);
                }
                else if (evt.nameEquals(SimList.OBJECT_REMOVED)) {
                    this.displayList.remove(dispSpring);
                }
            }
        }, 'add/remove spring during mouse drag');
        new GenericObserver(this.simRun, (evt) => {
            if (evt.nameEquals(SimRunner.en.RUNNING)) {
                const bp = evt;
                sim.setRunning(bp.getValue());
                if (bp.getValue() && sim.getTime() == 0) {
                    sim.saveInitialState();
                }
            }
        }, 'save initial conditions when starting to run at time 0');
        this.addPlaybackControls();
        let pn = sim.getParameterNumber(MoveablePendulumSim.en.LENGTH);
        this.addControl(new SliderControl(pn, 0.1, 10.1, true));
        pn = sim.getParameterNumber(MoveablePendulumSim.en.DAMPING);
        this.addControl(new SliderControl(pn, 0, 10, false));
        pn = sim.getParameterNumber(MoveablePendulumSim.en.MASS);
        this.addControl(new SliderControl(pn, 0.2, 20.2, true));
        pn = sim.getParameterNumber(MoveablePendulumSim.en.GRAVITY);
        this.addControl(new SliderControl(pn, 0, 20, false));
        pn = sim.getParameterNumber(MoveablePendulumSim.en.DRIVE_AMPLITUDE);
        this.addControl(new SliderControl(pn, 0, 400, false));
        pn = sim.getParameterNumber(MoveablePendulumSim.en.DRIVE_FREQUENCY);
        this.addControl(new SliderControl(pn, 0, 100, false));
        pn = sim.getParameterNumber(EnergyInfo.en.PE_OFFSET);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript();
        this.addURLScriptButton();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', anchor: ' + this.anchor.toStringShort()
            + ', bob: ' + this.bob.toStringShort()
            + ', rod: ' + this.rod.toStringShort()
            + super.toString();
    }
    ;
    getClassName() {
        return 'MoveablePendulumApp';
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('rod|anchor|bob', myName + '.');
    }
    ;
}
Util.defineGlobal('sims$pendulum$MoveablePendulumApp', MoveablePendulumApp);
