import { AbstractODESim } from '../../lab/model/ODESim.js';
import { ConcreteLine } from '../../lab/model/ConcreteLine.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const ANGLE = 0;
const ANGLEP = 1;
const TIME = 2;
const ANCHORX = 3;
const ANCHORXP = 4;
const ANCHORY = 5;
const ANCHORYP = 6;
const KE = 7;
const PE = 8;
const TE = 9;
export class MoveablePendulumSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.length_ = 1;
        this.gravity_ = 10;
        this.damping_ = 0.5;
        this.pendulumDragging_ = false;
        this.springDragging_ = false;
        this.anchorDamping_ = 0.8;
        this.springStiffness_ = 3;
        this.frequency_ = 30;
        this.amplitude_ = 200;
        this.potentialOffset_ = 0;
        this.running_ = false;
        const var_names = [
            MoveablePendulumSim.en.ANGLE,
            MoveablePendulumSim.en.ANGULAR_VELOCITY,
            VarsList.en.TIME,
            MoveablePendulumSim.en.ANCHOR_X,
            MoveablePendulumSim.en.ANCHOR_X_VELOCITY,
            MoveablePendulumSim.en.ANCHOR_Y,
            MoveablePendulumSim.en.ANCHOR_Y_VELOCITY,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY
        ];
        const i18n_names = [
            MoveablePendulumSim.i18n.ANGLE,
            MoveablePendulumSim.i18n.ANGULAR_VELOCITY,
            VarsList.i18n.TIME,
            MoveablePendulumSim.i18n.ANCHOR_X,
            MoveablePendulumSim.i18n.ANCHOR_X_VELOCITY,
            MoveablePendulumSim.i18n.ANCHOR_Y,
            MoveablePendulumSim.i18n.ANCHOR_Y_VELOCITY,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.anchor_ = PointMass.makeSquare(0.3, 'anchor');
        this.anchor_.setMass(1.0);
        this.anchor_.setPosition(Vector.ORIGIN);
        this.rod_ = new ConcreteLine('rod');
        this.bob_ = PointMass.makeCircle(0.2, 'bob');
        this.bob_.setMass(1);
        this.mouse_ = PointMass.makeCircle(0.2, 'mouse');
        this.mouse_.setMass(1.0);
        this.dragSpring_ = new Spring('dragSpring', this.mouse_, Vector.ORIGIN, this.anchor_, Vector.ORIGIN, 0, this.springStiffness_);
        this.getSimList().add(this.anchor_, this.bob_, this.rod_);
        this.getVarsList().setValue(ANGLE, Math.PI * 0.95);
        this.saveInitialState();
        this.setAnchorYVelocity();
        this.modifyObjects();
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.LENGTH, MoveablePendulumSim.i18n.LENGTH, () => this.getLength(), a => this.setLength(a)));
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.DAMPING, MoveablePendulumSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.MASS, MoveablePendulumSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.GRAVITY, MoveablePendulumSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.DRIVE_AMPLITUDE, MoveablePendulumSim.i18n.DRIVE_AMPLITUDE, () => this.getDriveAmplitude(), a => this.setDriveAmplitude(a)));
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.DRIVE_FREQUENCY, MoveablePendulumSim.i18n.DRIVE_FREQUENCY, () => this.getDriveFrequency(), a => this.setDriveFrequency(a)));
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.ANCHOR_DAMPING, MoveablePendulumSim.i18n.ANCHOR_DAMPING, () => this.getAnchorDamping(), a => this.setAnchorDamping(a)));
        this.addParameter(new ParameterNumber(this, MoveablePendulumSim.en.SPRING_STIFFNESS, MoveablePendulumSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        let pn;
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', length_: ' + Util.NF(this.length_)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', damping_: ' + Util.NF(this.damping_)
            + ', frequency_: ' + Util.NF(this.frequency_)
            + ', amplitude_: ' + Util.NF(this.amplitude_)
            + ', anchorDamping_: ' + Util.NF(this.anchorDamping_)
            + ', springStiffness_: ' + Util.NF(this.springStiffness_)
            + ', anchor_: ' + this.anchor_
            + ', bob_: ' + this.bob_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'MoveablePendulumSim';
    }
    ;
    setRunning(value) {
        this.running_ = value;
    }
    ;
    setAnchorYVelocity() {
        const value = Math.abs(this.frequency_) < 1E-10 ? 0 :
            -this.amplitude_ / this.frequency_;
        if (this.initialState_) {
            this.initialState_[ANCHORYP] = value * Math.cos(this.frequency_ * this.initialState_[TIME]);
        }
        const va = this.getVarsList();
        va.setValue(ANCHORYP, value * Math.cos(this.frequency_ * this.getTime()));
    }
    ;
    getEnergyInfo() {
        const ke = this.bob_.getKineticEnergy();
        const anchorY = this.anchor_.getPosition().getY();
        const y = this.bob_.getPosition().getY();
        const pe = this.gravity_ * this.bob_.getMass() * (y - anchorY + this.length_);
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        const angle = vars[ANGLE];
        const sinAngle = Math.sin(angle);
        const cosAngle = Math.cos(angle);
        this.anchor_.setPosition(new Vector(vars[ANCHORX], vars[ANCHORY]));
        const len = this.length_;
        this.bob_.setPosition(new Vector(vars[ANCHORX] + len * sinAngle, vars[ANCHORY] - len * cosAngle));
        const vx = vars[ANGLEP] * len * cosAngle;
        const vy = vars[ANGLEP] * len * sinAngle;
        this.bob_.setVelocity(new Vector(vx, vy));
        this.rod_.setStartPoint(this.anchor_.getPosition());
        this.rod_.setEndPoint(this.bob_.getPosition());
    }
    ;
    getDragSpring() {
        return this.dragSpring_;
    }
    ;
    startDrag(simObject, location, offset, _dragBody, _modifiers) {
        if (simObject == this.anchor_) {
            this.springDragging_ = this.running_;
            if (this.springDragging_) {
                this.getSimList().add(this.dragSpring_);
            }
            this.mouseDrag(simObject, location, offset);
            return true;
        }
        else if (simObject == this.bob_) {
            this.pendulumDragging_ = true;
            return true;
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const vars = va.getValues();
        const p = location.subtract(offset);
        if (simObject == this.anchor_) {
            if (this.springDragging_) {
                this.mouse_.setPosition(location);
            }
            else {
                va.setValue(ANCHORX, p.getX());
                va.setValue(ANCHORY, p.getY());
            }
        }
        else if (simObject == this.bob_) {
            const th = Math.PI / 2 + Math.atan2(p.getY() - vars[ANCHORY], p.getX() - vars[ANCHORX]);
            va.setValue(ANGLE, th);
            va.setValue(ANGLEP, 0);
        }
        this.moveObjects(va.getValues());
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.pendulumDragging_ = false;
        if (this.springDragging_) {
            this.springDragging_ = false;
            this.getSimList().remove(this.dragSpring_);
        }
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        this.moveObjects(vars);
        change[TIME] = 1;
        change[ANCHORX] = vars[ANCHORXP];
        change[ANCHORY] = vars[ANCHORYP];
        change[ANCHORXP] = -this.anchorDamping_ * vars[ANCHORXP];
        const mouse = this.mouse_.getPosition();
        if (this.springDragging_) {
            change[ANCHORXP] += this.springStiffness_ * (mouse.getX() - vars[ANCHORX]);
        }
        change[ANCHORYP] = -this.anchorDamping_ * vars[ANCHORYP] +
            this.amplitude_ * Math.sin(this.frequency_ * vars[TIME]);
        if (this.springDragging_) {
            change[ANCHORYP] += this.springStiffness_ * (mouse.getY() - vars[ANCHORY]);
        }
        if (!this.pendulumDragging_) {
            change[ANGLE] = vars[ANGLEP];
            const ddx0 = change[ANCHORXP];
            const ddy0 = change[ANCHORYP];
            const R = this.length_;
            let dd = -(this.gravity_ / R) * Math.sin(vars[ANGLE]);
            const mRsq = this.bob_.getMass() * R * R;
            dd += -(this.damping_ / mRsq) * vars[ANGLEP];
            dd += -(ddx0 / R) * Math.cos(vars[ANGLE]) - (ddy0 / R) * Math.sin(vars[ANGLE]);
            change[ANGLEP] = dd;
        }
        return null;
    }
    ;
    getMass() {
        return this.bob_.getMass();
    }
    ;
    setMass(value) {
        this.bob_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(MoveablePendulumSim.en.MASS);
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(MoveablePendulumSim.en.GRAVITY);
    }
    ;
    getDriveFrequency() {
        return this.frequency_;
    }
    ;
    setDriveFrequency(value) {
        this.frequency_ = value;
        this.setAnchorYVelocity();
        this.broadcastParameter(MoveablePendulumSim.en.DRIVE_FREQUENCY);
    }
    ;
    getDriveAmplitude() {
        return this.amplitude_;
    }
    ;
    setDriveAmplitude(value) {
        this.amplitude_ = value;
        this.setAnchorYVelocity();
        this.broadcastParameter(MoveablePendulumSim.en.DRIVE_AMPLITUDE);
    }
    ;
    getLength() {
        return this.length_;
    }
    ;
    setLength(value) {
        this.length_ = value;
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(MoveablePendulumSim.en.LENGTH);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(MoveablePendulumSim.en.DAMPING);
    }
    ;
    getSpringStiffness() {
        return this.springStiffness_;
    }
    ;
    setSpringStiffness(value) {
        this.springStiffness_ = value;
        this.dragSpring_.setStiffness(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(MoveablePendulumSim.en.SPRING_STIFFNESS);
    }
    ;
    getAnchorDamping() {
        return this.anchorDamping_;
    }
    ;
    setAnchorDamping(value) {
        this.anchorDamping_ = value;
        this.broadcastParameter(MoveablePendulumSim.en.ANCHOR_DAMPING);
    }
    ;
}
MoveablePendulumSim.en = {
    DRIVE_AMPLITUDE: 'drive amplitude',
    ANGLE: 'angle',
    ANGULAR_VELOCITY: 'angle velocity',
    DAMPING: 'damping',
    DRIVE_FREQUENCY: 'drive frequency',
    GRAVITY: 'gravity',
    LENGTH: 'length',
    MASS: 'mass',
    SPRING_STIFFNESS: 'spring stiffness',
    ANCHOR_DAMPING: 'anchor damping',
    ANCHOR_X: 'anchor X',
    ANCHOR_X_VELOCITY: 'anchor X velocity',
    ANCHOR_Y: 'anchor Y',
    ANCHOR_Y_VELOCITY: 'anchor Y velocity'
};
MoveablePendulumSim.de_strings = {
    DRIVE_AMPLITUDE: 'Antriebsamplitude',
    ANGLE: 'Winkel',
    ANGULAR_VELOCITY: 'Winkel Geschwindigkeit',
    DAMPING: 'Dämpfung',
    DRIVE_FREQUENCY: 'Antriebsfrequenz',
    GRAVITY: 'Gravitation',
    LENGTH: 'Länge',
    MASS: 'Masse',
    SPRING_STIFFNESS: 'Federsteifheit',
    ANCHOR_DAMPING: 'Anker Dämpfung',
    ANCHOR_X: 'Anker X',
    ANCHOR_X_VELOCITY: 'Anker X Geschwindigkeit',
    ANCHOR_Y: 'Anker Y',
    ANCHOR_Y_VELOCITY: 'Anker Y Geschwindigkeit'
};
MoveablePendulumSim.i18n = Util.LOCALE === 'de' ? MoveablePendulumSim.de_strings : MoveablePendulumSim.en;
Util.defineGlobal('sims$pendulum$MoveablePendulumSim', MoveablePendulumSim);
