import { AbstractApp } from '../common/AbstractApp.js';
import { BrachistoObserver } from './BrachistoObserver.js';
import { BrachistochronePath, LinearPath, CircleArcPath, Brachistochrone2Path, ParabolaUpPath, ParabolaDownPath } from './BrachistoPaths.js';
import { BrachistoSim } from './BrachistoSim.js';
import { ClockTask } from '../../lab/util/Clock.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { GenericObserver } from '../../lab/util/Observe.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { SimpleAdvance } from '../../lab/model/SimpleAdvance.js';
import { SimRunner } from '../../lab/app/SimRunner.js';
import { Util } from '../../lab/util/Util.js';
export class BrachistoApp extends AbstractApp {
    constructor(elem_ids) {
        Util.setErrorHandler();
        const paths = [
            new BrachistochronePath(),
            new LinearPath(),
            new CircleArcPath(),
            new Brachistochrone2Path(),
            new ParabolaUpPath(),
            new ParabolaDownPath()
        ];
        const simRect = new DoubleRect(-1, -3, 7, 1);
        const sim = new BrachistoSim(paths);
        const advance = new SimpleAdvance(sim);
        super(elem_ids, simRect, sim, advance, sim, null);
        this.paths = paths;
        this.brachistoObserver = new BrachistoObserver(sim, this.simList, this.simView, this.statusView);
        new GenericObserver(sim, (evt) => {
            if (evt.nameEquals(BrachistoSim.PATH_CHOSEN)) {
                this.clock.setTime(0);
                this.clock.setRealTime(0);
                const choice = evt.getValue();
                if (choice == -1) {
                    this.clock.pause();
                }
                else {
                    this.clock.resume();
                }
            }
        }, 'start clock when path chosen');
        new GenericObserver(this.simRun, (evt) => {
            if (evt.nameEquals(SimRunner.RESET)) {
                sim.setPathChoice(-1);
            }
        }, 'reset path choice when reset occurs');
        sim.setPathChoice(-1);
        this.task = this.makeTask(6);
        this.clock.addTask(this.task);
        this.addPlaybackControls();
        let pn = sim.getParameterNumber(BrachistoSim.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(BrachistoSim.en.DAMPING);
        this.addControl(new NumericControl(pn));
        pn = sim.getParameterNumber(BrachistoSim.en.MASS);
        this.addControl(new NumericControl(pn));
        this.addParameter(pn = new ParameterNumber(this, BrachistoApp.en.REPEAT_TIME, BrachistoApp.i18n.REPEAT_TIME, () => this.getRepeatTime(), a => this.setRepeatTime(a)));
        pn.setSignifDigits(0);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.graph.line.setXVariable(0);
        this.graph.line.setYVariable(1);
        this.timeGraph.line1.setYVariable(0);
        this.timeGraph.line2.setYVariable(1);
        this.makeEasyScript();
        this.addURLScriptButton();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', brachistoObserver: ' + this.brachistoObserver.toStringShort()
            + ', paths: [ ' + this.paths + ' ]'
            + super.toString();
    }
    ;
    getClassName() {
        return 'BrachistoApp';
    }
    ;
    getRepeatTime() {
        return this.task.getTime();
    }
    ;
    makeTask(time) {
        return new ClockTask(time, () => {
            this.sim.reset();
            this.clock.setTime(0);
            this.clock.setRealTime(0);
        });
    }
    ;
    setRepeatTime(value) {
        this.clock.removeTask(this.task);
        this.task = this.makeTask(value);
        this.clock.addTask(this.task);
        this.sim.reset();
        this.clock.setTime(0);
        this.clock.setRealTime(0);
        this.broadcastParameter(BrachistoApp.en.REPEAT_TIME);
    }
    ;
    defineNames(myName) {
        super.defineNames(myName);
        this.terminal.addRegex('brachistoObserver|paths', myName + '.');
        this.terminal.addRegex('LinearPath|BrachistochronePath|' +
            'ParabolaUpPath|Brachistochrone2Path|ParabolaDownPath|CircleArcPath', 'sims$roller$');
    }
    ;
    setup() {
        super.setup();
        this.clock.pause();
    }
    ;
}
BrachistoApp.en = {
    REPEAT_TIME: 'repeat time'
};
BrachistoApp.de_strings = {
    REPEAT_TIME: 'Intervallwiederholung'
};
BrachistoApp.i18n = Util.LOCALE === 'de' ? BrachistoApp.de_strings : BrachistoApp.en;
Util.defineGlobal('sims$roller$BrachistoApp', BrachistoApp);
