import { CardioidPath } from './CardioidPath.js';
import { ChoiceControl } from '../../lab/controls/ChoiceControl.js';
import { CirclePath } from './CirclePath.js';
import { CollisionAdvance } from '../../lab/model/CollisionAdvance.js';
import { CommonControls } from '../common/CommonControls.js';
import { ContactSim } from '../../lab/engine2D/ContactSim.js';
import { DampingLaw } from '../../lab/model/DampingLaw.js';
import { DoubleRect } from '../../lab/util/DoubleRect.js';
import { Engine2DApp } from '../engine2D/Engine2DApp.js';
import { FlatPath } from './FlatPath.js';
import { FunctionVariable } from '../../lab/model/FunctionVariable.js';
import { GenericMemo } from '../../lab/util/Memo.js';
import { GenericObserver } from '../../lab/util/Observe.js';
import { GravityLaw } from '../../lab/model/GravityLaw.js';
import { HumpPath } from './HumpPath.js';
import { LemniscatePath } from './LemniscatePath.js';
import { LoopTheLoopPath } from './LoopTheLoopPath.js';
import { NumericalPath } from '../../lab/model/NumericalPath.js';
import { NumericControl } from '../../lab/controls/NumericControl.js';
import { OvalPath } from './OvalPath.js';
import { PathEndPoint } from '../../lab/engine2D/PathEndPoint.js';
import { PathJoint } from '../../lab/engine2D/PathJoint.js';
import { PathObserver } from './PathObserver.js';
import { PathSelector } from './PathSelector.js';
import { Scrim } from '../../lab/engine2D/Scrim.js';
import { Shapes } from '../../lab/engine2D/Shapes.js';
import { SpiralPath } from './SpiralPath.js';
import { Util } from '../../lab/util/Util.js';
import { Vector } from '../../lab/util/Vector.js';
export class RigidBodyRollerApp extends Engine2DApp {
    constructor(elem_ids) {
        const simRect = new DoubleRect(-6, -6, 6, 6);
        const sim = new ContactSim();
        const advance = new CollisionAdvance(sim);
        super(elem_ids, simRect, sim, advance);
        this.pathAction = null;
        this.resetObserver = null;
        this.layout.getSimCanvas().setBackground('black');
        this.layout.getSimCanvas().setAlpha(CommonControls.SHORT_TRAILS);
        this.elasticity.setElasticity(0.8);
        this.sim.setShowForces(true);
        this.sim.setExtraAccel("velocity_joints");
        this.dampingLaw = new DampingLaw(0, 0.15, this.simList);
        this.gravityLaw = new GravityLaw(3, this.simList);
        this.paths = [
            new HumpPath(),
            new LoopTheLoopPath(),
            new CirclePath(3.0),
            new OvalPath(),
            new LemniscatePath(2.0),
            new CardioidPath(3.0),
            new SpiralPath(),
            new FlatPath()
        ];
        this.path = new NumericalPath(this.paths[2]);
        this.pathSelect = new PathSelector(this, this.paths);
        this.pathObserver = new PathObserver(this.simList, this.simView, a => this.setSimRect(a), 1.5);
        this.block = Shapes.makeBlock(1, 3, RigidBodyRollerApp.en.BLOCK, RigidBodyRollerApp.i18n.BLOCK);
        this.addPlaybackControls();
        let ps = this.pathSelect.getParameterString(PathSelector.en.PATH);
        this.addControl(new ChoiceControl(ps));
        let pn = this.gravityLaw.getParameterNumber(GravityLaw.en.GRAVITY);
        this.addControl(new NumericControl(pn));
        this.watchEnergyChange(pn);
        pn = this.dampingLaw.getParameterNumber(DampingLaw.en.DAMPING);
        this.addControl(new NumericControl(pn));
        this.addStandardControls();
        this.makeEasyScript([this.simView]);
        this.addURLScriptButton();
        this.setPath(this.path);
        this.graphSetup();
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', simList: ' + this.simList.toStringShort()
            + ', simView: ' + this.simView.toStringShort()
            + ', statusView: ' + this.statusView.toStringShort()
            + ', pathSelect: ' + this.pathSelect
            + ', pathObserver: ' + this.pathObserver
            + ', paths: [ ' + this.paths + ' ]'
            + super.toString();
    }
    ;
    getClassName() {
        return 'RigidBodyRollerApp';
    }
    ;
    defineNames(myName) {
        this.terminal.addRegex('block|paths|path|pathSelect|gravityLaw|dampingLaw', myName + '.');
        this.terminal.addRegex('CardioidPath|CirclePath'
            + '|FlatPath|HumpPath|LemniscatePath|LoopTheLoopPath|OvalPath'
            + '|PathObserver|PathSelector|SpiralPath', 'sims$roller$', false);
    }
    ;
    getSubjects() {
        return super.getSubjects().concat(this.pathSelect, this.gravityLaw, this.dampingLaw);
    }
    ;
    graphSetup(_body) {
        this.graph.line.setXVariable(10);
        this.graph.line.setYVariable(1);
        this.timeGraph.line1.setYVariable(1);
        this.timeGraph.line2.setYVariable(2);
    }
    ;
    config() {
        Util.assert(this.path != null);
        if (this.resetObserver != null) {
            this.resetObserver.disconnect();
            this.resetObserver = null;
        }
        if (this.pathAction != null) {
            this.simRun.removeMemo(this.pathAction);
            this.pathAction = null;
        }
        const elasticity = this.elasticity.getElasticity();
        this.sim.cleanSlate();
        this.advance.reset();
        this.sim.addForceLaw(this.gravityLaw);
        this.gravityLaw.connect(this.sim.getSimList());
        this.sim.addForceLaw(this.dampingLaw);
        this.dampingLaw.connect(this.sim.getSimList());
        this.simList.add(this.path);
        this.block.setPosition(new Vector(-4, 4), Math.PI / 4);
        this.block.setVelocity(new Vector(0, 0), 0);
        this.sim.addBody(this.block);
        let ds = this.displayList.findShape(this.block);
        ds.setFillStyle('rgba(51,204,255,0.5)');
        ds.setDrawCenterOfMass(true);
        ds.setDrawDragPoints(true);
        const attach = this.block.getDragPoints()[1];
        this.pathJoint = new PathJoint(this.path, this.block, attach);
        this.sim.addConnector(this.pathJoint);
        this.sim.alignConnectors();
        if (this.path.nameEquals(CirclePath.en.NAME)) {
            if (0 == 1) {
                const endPt = this.path.findNearestGlobal(new Vector(-2, -2));
                this.sim.addConnector(new PathEndPoint('stop point', this.path, this.block, attach, endPt.p, true));
            }
            this.sim.addConnector(new PathEndPoint('limit down', this.path, this.block, attach, this.path.getStartPValue(), false));
            this.sim.addConnector(new PathEndPoint('limit up', this.path, this.block, attach, this.path.getFinishPValue(), true));
        }
        else if (this.path.nameEquals(HumpPath.en.NAME)) {
            this.sim.addConnector(new PathEndPoint('limit down', this.path, this.block, attach, this.path.getStartPValue() + 0.1, false));
            this.sim.addConnector(new PathEndPoint('limit up', this.path, this.block, attach, this.path.getFinishPValue() - 0.1, true));
            const disconnectPt = 3 * this.path.getFinishPValue() / 4;
            this.pathAction = new GenericMemo(() => {
                const ppt = this.pathJoint.getPathPoint();
                if (ppt.p > disconnectPt) {
                    this.sim.removeConnector(this.pathJoint);
                }
            }, 'disconnect block from path');
            this.simRun.addMemo(this.pathAction);
            const dpt = this.path.map_p_to_vector(disconnectPt);
            const dj = new PathJoint(this.path, Scrim.getScrim(), dpt);
            this.sim.getSimList().add(dj);
            this.resetObserver = new GenericObserver(this.sim, (evt) => {
                if (evt.nameEquals('RESET')) {
                    this.sim.addConnector(this.pathJoint);
                    this.sim.alignConnectors();
                }
            }, 'Add back PathJoint on RESET event');
        }
        const va = this.sim.getVarsList();
        const varP = new FunctionVariable(va, RigidBodyRollerApp.en.PATH_POSITION, RigidBodyRollerApp.i18n.PATH_POSITION, () => this.pathJoint.getPathPoint().p);
        va.addVariable(varP);
        const varPV = new FunctionVariable(va, RigidBodyRollerApp.en.PATH_VELOCITY, RigidBodyRollerApp.i18n.PATH_VELOCITY, () => {
            const ppt = this.pathJoint.getPathPoint();
            const vel = this.block.getVelocity(this.pathJoint.getAttach1());
            return vel.dotProduct(new Vector(ppt.slopeX, ppt.slopeY));
        });
        va.addVariable(varPV);
        this.sim.setElasticity(elasticity);
        this.sim.saveInitialState();
        this.easyScript.update();
    }
    ;
    getPath() {
        return this.path;
    }
    ;
    setPath(path) {
        this.path = path;
        this.config();
    }
    ;
    setSimRect(simRect) {
        this.simRect = simRect;
        this.simView.setSimRect(simRect);
    }
    ;
}
RigidBodyRollerApp.en = {
    BLOCK: 'block',
    PATH_POSITION: 'path position',
    PATH_VELOCITY: 'path velocity'
};
RigidBodyRollerApp.de_strings = {
    BLOCK: 'Block',
    PATH_POSITION: 'Pfad Position',
    PATH_VELOCITY: 'Pfad Geschwindigkeit'
};
RigidBodyRollerApp.i18n = Util.LOCALE === 'de' ? RigidBodyRollerApp.de_strings : RigidBodyRollerApp.en;
Util.defineGlobal('sims$roller$RigidBodyRollerApp', RigidBodyRollerApp);
