import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PathPoint } from '../../lab/model/PathPoint.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { RollerCollision } from './RollerCollision.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const TRACK_P = 0;
const TRACK_V = 1;
const X = 2;
const Y = 3;
const XP = 4;
const YP = 5;
const TRACK_MODE = 6;
const KE = 7;
const PE = 8;
const TE = 9;
const TIME = 10;
const ANCHOR_X = 11;
const ANCHOR_Y = 12;
const ON_TRACK = 1;
const OFF_TRACK = 0;
export class RollerFlightSim extends AbstractODESim {
    constructor(thePath, opt_name) {
        super(opt_name);
        this.damping_ = 0;
        this.gravity_ = 9.8;
        this.stickiness_ = 0.1;
        this.elasticity_ = 0.8;
        this.potentialOffset_ = 0;
        this.debugPaint_ = null;
        this.dragObj_ = null;
        this.collisionFunction_ = null;
        const var_names = [
            RollerFlightSim.en.TRACK_POSITION,
            RollerFlightSim.en.TRACK_VELOCITY,
            RollerFlightSim.en.X_POSITION,
            RollerFlightSim.en.Y_POSITION,
            RollerFlightSim.en.X_VELOCITY,
            RollerFlightSim.en.Y_VELOCITY,
            RollerFlightSim.en.TRACK_MODE,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
            VarsList.en.TIME,
            RollerFlightSim.en.ANCHOR_X,
            RollerFlightSim.en.ANCHOR_Y
        ];
        const i18n_names = [
            RollerFlightSim.i18n.TRACK_POSITION,
            RollerFlightSim.i18n.TRACK_VELOCITY,
            RollerFlightSim.i18n.X_POSITION,
            RollerFlightSim.i18n.Y_POSITION,
            RollerFlightSim.i18n.X_VELOCITY,
            RollerFlightSim.i18n.Y_VELOCITY,
            RollerFlightSim.i18n.TRACK_MODE,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
            VarsList.i18n.TIME,
            RollerFlightSim.i18n.ANCHOR_X,
            RollerFlightSim.i18n.ANCHOR_Y
        ];
        this.setVarsList(new VarsList(var_names, i18n_names, this.getName() + '_VARS'));
        this.getVarsList().setComputed(KE, PE, TE);
        this.getVarsList().getVariable(TRACK_MODE).setBroadcast(true);
        this.ball1_ = PointMass.makeCircle(0.3, 'ball1');
        this.ball1_.setMass(0.5);
        this.anchor_ = PointMass.makeSquare(0.4, 'anchor');
        this.spring_ = new Spring('spring', this.ball1_, Vector.ORIGIN, this.anchor_, Vector.ORIGIN, 1.0, 0);
        this.path_ = thePath;
        const r = this.path_.getBoundsWorld();
        this.lowestPoint_ = r.getBottom();
        const start = new Vector(r.getLeft() + r.getWidth() * 0.1, r.getTop() - r.getHeight() * 0.1);
        const pathPoint = this.path_.findNearestGlobal(start);
        this.xLow_ = this.path_.map_p_to_x(this.path_.getStartPValue());
        this.xHigh_ = this.path_.map_p_to_x(this.path_.getFinishPValue());
        const va = this.getVarsList();
        va.setValue(TRACK_P, pathPoint.p);
        va.setValue(TRACK_V, 0);
        va.setValue(TRACK_MODE, ON_TRACK);
        va.setValue(ANCHOR_X, r.getLeft() + r.getWidth() * 0.2);
        va.setValue(ANCHOR_Y, r.getTop() - r.getHeight() * 0.5);
        this.modifyObjects();
        this.saveInitialState();
        this.getSimList().add(this.path_, this.ball1_, this.anchor_, this.spring_);
        let pn;
        this.addParameter(pn = new ParameterNumber(this, RollerFlightSim.en.STICKINESS, RollerFlightSim.i18n.STICKINESS, () => this.getStickiness(), a => this.setStickiness(a)));
        pn.setSignifDigits(3);
        this.addParameter(pn = new ParameterNumber(this, RollerFlightSim.en.ELASTICITY, RollerFlightSim.i18n.ELASTICITY, () => this.getElasticity(), a => this.setElasticity(a)));
        pn.setSignifDigits(3);
        pn.setUpperLimit(1);
        this.addParameter(new ParameterNumber(this, RollerFlightSim.en.DAMPING, RollerFlightSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        this.addParameter(new ParameterNumber(this, RollerFlightSim.en.GRAVITY, RollerFlightSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, RollerFlightSim.en.MASS, RollerFlightSim.i18n.MASS, () => this.getMass(), a => this.setMass(a)));
        this.addParameter(new ParameterNumber(this, RollerFlightSim.en.SPRING_LENGTH, RollerFlightSim.i18n.SPRING_LENGTH, () => this.getSpringLength(), a => this.setSpringLength(a)));
        this.addParameter(new ParameterNumber(this, RollerFlightSim.en.SPRING_STIFFNESS, RollerFlightSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', ball1: ' + this.ball1_
            + ', anchor: ' + this.anchor_
            + ', spring: ' + this.spring_
            + ', path: ' + this.path_
            + ', damping: ' + Util.NF(this.damping_)
            + ', gravity: ' + Util.NF(this.gravity_)
            + ', stickiness: ' + Util.NF(this.stickiness_)
            + ', elasticity: ' + Util.NF(this.elasticity_)
            + ', lowestPoint: ' + Util.NF(this.lowestPoint_)
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'RollerFlightSim';
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        const currentPoint = this.moveObjects(vars);
        if (vars[TRACK_MODE] == ON_TRACK) {
            va.setValue(X, this.ball1_.getPosition().getX(), true);
            va.setValue(Y, this.ball1_.getPosition().getY(), true);
            va.setValue(XP, this.ball1_.getVelocity().getX(), true);
            va.setValue(YP, this.ball1_.getVelocity().getY(), true);
        }
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
        if (vars[TRACK_MODE] == ON_TRACK) {
            if (currentPoint != null) {
                this.jumpOffTrack(currentPoint);
            }
            else {
                throw ('advance: currentPoint is null, TRACK_MODE=' + vars[TRACK_MODE]);
            }
        }
    }
    ;
    setDebugPaint(fn) {
        this.debugPaint_ = fn;
    }
    ;
    moveObjects(vars) {
        let pathPoint = null;
        this.anchor_.setPosition(new Vector(vars[ANCHOR_X], vars[ANCHOR_Y]));
        if (vars[TRACK_MODE] == ON_TRACK) {
            pathPoint = new PathPoint(vars[TRACK_P], true);
            this.path_.map_p_to_slope(pathPoint);
            this.ball1_.setPosition(pathPoint);
            this.ball1_.setVelocity(new Vector(pathPoint.slopeX * vars[TRACK_V], pathPoint.slopeY * vars[TRACK_V]));
        }
        else if (vars[TRACK_MODE] == OFF_TRACK) {
            this.ball1_.setPosition(new Vector(vars[X], vars[Y]));
            this.ball1_.setVelocity(new Vector(vars[XP], vars[YP], 0));
        }
        else {
            throw ('moveObjects: TRACK_MODE=' + vars[TRACK_MODE]);
        }
        if (this.debugPaint_ != null) {
            this.debugPaint_();
        }
        return pathPoint;
    }
    jumpOffTrack(pathPoint1) {
        const va = this.getVarsList();
        Util.assert(va.getValue(TRACK_MODE) == ON_TRACK);
        const r = pathPoint1.radius;
        const direction = pathPoint1.direction;
        const k = pathPoint1.slope;
        const slopeDenom = Math.sqrt(1 + k * k);
        let g = this.gravity_ / slopeDenom;
        if (r > 0) {
            g = -g;
        }
        let ar = g;
        if (this.spring_.getStiffness() > 0) {
            const x = pathPoint1.getX();
            const y = pathPoint1.getY();
            const b = y - k * x;
            const p2 = this.spring_.getEndPoint();
            const below = (p2.getY() < k * p2.getX() + b) ? 1 : -1;
            const sv = this.spring_.getVector();
            let costh = direction * (sv.getX() + k * sv.getY()) / this.spring_.getLength();
            costh = costh / slopeDenom;
            const sinth = Math.sqrt(1 - costh * costh);
            if (isNaN(sinth) || sinth > 1 || sinth < 0) {
                throw 'sin(theta) out of range ' + sinth;
            }
            const as = (sinth * this.spring_.getStretch() * this.spring_.getStiffness()) /
                this.ball1_.getMass();
            if (r < 0) {
                ar = ar + below * as;
            }
            else {
                ar = ar - below * as;
            }
        }
        const v = va.getValue(1);
        const av = Math.abs(v * v / r);
        if (r < 0 && ar < av || r > 0 && ar > av) {
            va.setValue(TRACK_MODE, OFF_TRACK);
            va.setValue(X, pathPoint1.getX(), true);
            va.setValue(Y, pathPoint1.getY(), true);
            const point2 = new PathPoint();
            point2.x = pathPoint1.getX();
            this.path_.map_x_to_y_p(point2);
            if (va.getValue(3) < point2.getY()) {
                va.setValue(Y, point2.getY());
            }
            va.setValue(XP, v / slopeDenom, true);
            va.setValue(YP, v * k / slopeDenom, true);
        }
    }
    ;
    getEnergyInfo() {
        const ke = this.ball1_.getKineticEnergy();
        let pe = this.ball1_.getMass() * this.gravity_ *
            (this.ball1_.getPosition().getY() - this.lowestPoint_);
        pe += this.spring_.getPotentialEnergy();
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.ball1_) {
            this.dragObj_ = simObject;
            return true;
        }
        else if (simObject == this.anchor_) {
            return true;
        }
        return false;
    }
    ;
    off_track(x) {
        if (this.path_.isClosedLoop())
            return false;
        else {
            return (x < this.xLow_ || x > this.xHigh_);
        }
    }
    off_track_adjust(x) {
        if (x < this.xLow_)
            x = this.xLow_ + 0.1;
        if (x > this.xHigh_)
            x = this.xHigh_ - 0.1;
        return x;
    }
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const p = location.subtract(offset);
        if (simObject == this.ball1_) {
            if (this.off_track(p.getX())) {
                const point1 = this.path_.findNearestGlobal(p);
                va.setValue(TRACK_P, point1.p);
                va.setValue(TRACK_V, 0);
                va.setValue(TRACK_MODE, ON_TRACK);
                va.incrSequence(X, Y, XP, YP, KE, PE, TE);
            }
            else {
                if (p.getY() < this.path_.map_x_to_y(p.getX())) {
                    const point2 = this.path_.findNearestGlobal(p);
                    va.setValue(TRACK_P, point2.p);
                    va.setValue(TRACK_V, 0);
                    va.setValue(TRACK_MODE, ON_TRACK);
                    va.incrSequence(X, Y, XP, YP, KE, PE, TE);
                }
                else {
                    va.setValue(X, p.getX());
                    va.setValue(Y, p.getY());
                    va.setValue(XP, 0);
                    va.setValue(YP, 0);
                    va.setValue(TRACK_MODE, OFF_TRACK);
                    va.incrSequence(TRACK_P, TRACK_V, KE, PE, TE);
                }
            }
        }
        else if (simObject == this.anchor_) {
            va.setValue(ANCHOR_X, p.getX());
            va.setValue(ANCHOR_Y, p.getY());
        }
        this.moveObjects(va.getValues());
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.dragObj_ = null;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        if (this.dragObj_ != null) {
            return null;
        }
        const stiffness = this.spring_.getStiffness();
        const stretch = this.spring_.getStretch();
        const mass = this.ball1_.getMass();
        if (vars[TRACK_MODE] == ON_TRACK) {
            const pathPoint = this.moveObjects(vars);
            if (pathPoint == null) {
                throw '';
            }
            change[TRACK_P] = vars[TRACK_V];
            const k = pathPoint.slope;
            const sinTheta = isFinite(k) ? k / Math.sqrt(1 + k * k) : 1;
            change[TRACK_V] = -this.gravity_ * pathPoint.direction * sinTheta
                - this.damping_ * vars[TRACK_V] / mass;
            if (stiffness > 0) {
                const sv = this.spring_.getVector();
                let cosTheta;
                if (!isFinite(k)) {
                    cosTheta = pathPoint.direction * sv.getY() / this.spring_.getLength();
                }
                else {
                    cosTheta = pathPoint.direction * (sv.getX() + k * sv.getY()) /
                        (this.spring_.getLength() * Math.sqrt(1 + k * k));
                }
                if (isNaN(cosTheta) || cosTheta > 1 || cosTheta < -1) {
                    throw 'cosTheta out of range ' + cosTheta;
                }
                change[TRACK_V] += stiffness * cosTheta * stretch / mass;
            }
        }
        else if (vars[TRACK_MODE] == OFF_TRACK) {
            this.moveObjects(vars);
            change[X] = vars[XP];
            change[Y] = vars[YP];
            if (stiffness > 0) {
                const sv = this.spring_.getVector().multiply(-1);
                const xx = sv.getX();
                const yy = sv.getY();
                const len = this.spring_.getLength();
                change[XP] = -(stiffness / mass) * stretch * xx / len;
                change[YP] = -(stiffness / mass) * stretch * yy / len;
            }
            change[YP] += -this.gravity_;
            change[XP] -= (this.damping_ / mass) * vars[XP];
            change[YP] -= (this.damping_ / mass) * vars[YP];
        }
        else {
            Util.assert(false);
        }
        return null;
    }
    ;
    findCollisions(collisions, vars, _stepSize) {
        this.moveObjects(vars);
        if (vars[TRACK_MODE] == OFF_TRACK) {
            const c = new RollerCollision(this.ball1_, this.path_, this.getTime());
            if (c.isColliding()) {
                collisions.push(c);
            }
        }
    }
    ;
    handleCollisions(collisions, opt_totals) {
        const va = this.getVarsList();
        const vars = va.getValues();
        if (collisions.length == 0 || vars[TRACK_MODE] == ON_TRACK) {
            return true;
        }
        const c = collisions[0];
        const m_Point = new PathPoint(c.getPathPoint().p);
        m_Point.idx = this.path_.map_p_to_index(m_Point.p);
        this.path_.findNearestLocal(new Vector(vars[X], vars[Y]), m_Point);
        this.path_.map_p_to_slope(m_Point);
        const k = m_Point.slope;
        if (this.off_track(vars[X])) {
            va.setValue(X, this.off_track_adjust(vars[X]));
            va.setValue(XP, 0);
            va.setValue(YP, 0);
        }
        else {
            const vx = vars[XP];
            const vy = vars[YP];
            const d = (vx + k * vy) / (1 + k * k);
            const cx = d;
            const cy = d * k;
            const nx = vx - cx;
            const ny = vy - cy;
            const rx = cx - this.elasticity_ * nx;
            const ry = cy - this.elasticity_ * ny;
            va.setValue(XP, rx);
            va.setValue(YP, ry);
            let nv = Math.sqrt(nx * nx + ny * ny);
            c.impulse = nv * (1 + this.elasticity_) * this.ball1_.getMass();
            nv = this.elasticity_ * nv;
            c.velocity = nv;
            const rv = Math.sqrt(rx * rx + ry * ry);
            if (nv / rv < this.stickiness_) {
                va.setValue(TRACK_P, m_Point.p);
                va.setValue(TRACK_V, Math.sqrt(cx * cx + cy * cy) * (cx > 0 ? 1 : -1));
                va.setValue(TRACK_MODE, ON_TRACK);
            }
        }
        if (opt_totals) {
            opt_totals.addImpulses(1);
        }
        if (this.collisionFunction_ && this.terminal_) {
            this.collisionFunction_(c, this.terminal_);
        }
        va.incrSequence(KE, PE, TE);
        return true;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerFlightSim.en.GRAVITY);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(RollerFlightSim.en.DAMPING);
    }
    ;
    getMass() {
        return this.ball1_.getMass();
    }
    ;
    setMass(value) {
        this.ball1_.setMass(value);
        this.getVarsList().incrSequence(KE, PE, TE);
        this.broadcastParameter(RollerFlightSim.en.MASS);
    }
    ;
    getSpringStiffness() {
        return this.spring_.getStiffness();
    }
    ;
    setSpringStiffness(value) {
        this.spring_.setStiffness(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerFlightSim.en.SPRING_STIFFNESS);
    }
    ;
    getSpringLength() {
        return this.spring_.getRestLength();
    }
    ;
    setSpringLength(value) {
        this.spring_.setRestLength(value);
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(RollerFlightSim.en.SPRING_LENGTH);
    }
    ;
    getElasticity() {
        return this.elasticity_;
    }
    ;
    setElasticity(value) {
        this.elasticity_ = value;
        this.broadcastParameter(RollerFlightSim.en.ELASTICITY);
    }
    ;
    getStickiness() {
        return this.stickiness_;
    }
    ;
    setStickiness(value) {
        let v = value;
        if (v < 0.001) {
            v = 0.001;
        }
        else if (v > 1) {
            v = 1;
        }
        this.stickiness_ = v;
        this.broadcastParameter(RollerFlightSim.en.STICKINESS);
    }
    ;
    setCollisionFunction(f) {
        this.collisionFunction_ = f;
    }
    ;
}
RollerFlightSim.en = {
    ANCHOR_X: 'anchor X',
    ANCHOR_Y: 'anchor Y',
    DAMPING: 'damping',
    ELASTICITY: 'elasticity',
    GRAVITY: 'gravity',
    MASS: 'mass',
    POSITION: 'position',
    SPRING_DAMPING: 'spring damping',
    SPRING_LENGTH: 'spring length',
    SPRING_STIFFNESS: 'spring stiffness',
    STICKINESS: 'stickiness',
    TRACK_MODE: 'track mode',
    TRACK_POSITION: 'track position',
    TRACK_VELOCITY: 'track velocity',
    VELOCITY: 'velocity',
    X_POSITION: 'x position',
    X_VELOCITY: 'x velocity',
    Y_POSITION: 'y position',
    Y_VELOCITY: 'y velocity'
};
RollerFlightSim.de_strings = {
    ANCHOR_X: 'Anker X',
    ANCHOR_Y: 'Anker Y',
    DAMPING: 'Dämpfung',
    ELASTICITY: 'Elastizität',
    GRAVITY: 'Gravitation',
    MASS: 'Masse',
    POSITION: 'Position',
    SPRING_DAMPING: 'Federdämpfung',
    SPRING_LENGTH: 'Federlänge',
    SPRING_STIFFNESS: 'Federsteifheit',
    STICKINESS: 'Klebrigkeit',
    TRACK_MODE: 'Spur mode',
    TRACK_POSITION: 'Spur Position',
    TRACK_VELOCITY: 'Spur Geschwindigkeit',
    VELOCITY: 'Geschwindigkeit',
    X_POSITION: 'x Position',
    X_VELOCITY: 'x Geschwindigkeit',
    Y_POSITION: 'y Position',
    Y_VELOCITY: 'y Geschwindigkeit'
};
RollerFlightSim.i18n = Util.LOCALE === 'de' ? RollerFlightSim.de_strings : RollerFlightSim.en;
Util.defineGlobal('sims$roller$RollerFlightSim', RollerFlightSim);
