import { AbstractODESim } from '../../lab/model/ODESim.js';
import { ConcreteLine } from '../../lab/model/ConcreteLine.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Spring } from '../../lab/model/Spring.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const THETA = 0;
const THETAP = 1;
const R = 2;
const RP = 3;
const PHI = 4;
const PHIP = 5;
const TIME = 6;
export class DangleStickSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.gravity_ = 9.8;
        this.stickLength_ = 1;
        this.stick_ = new ConcreteLine('stick');
        this.isDragging_ = false;
        const var_names = [
            DangleStickSim.en.SPRING_ANGLE,
            DangleStickSim.en.SPRING_ANGULAR_VELOCITY,
            DangleStickSim.en.SPRING_LENGTH,
            DangleStickSim.en.SPRING_LENGTH_VELOCITY,
            DangleStickSim.en.STICK_ANGLE,
            DangleStickSim.en.STICK_ANGULAR_VELOCITY,
            VarsList.en.TIME
        ];
        const i18n_names = [
            DangleStickSim.i18n.SPRING_ANGLE,
            DangleStickSim.i18n.SPRING_ANGULAR_VELOCITY,
            DangleStickSim.i18n.SPRING_LENGTH,
            DangleStickSim.i18n.SPRING_LENGTH_VELOCITY,
            DangleStickSim.i18n.STICK_ANGLE,
            DangleStickSim.i18n.STICK_ANGULAR_VELOCITY,
            VarsList.i18n.TIME
        ];
        const va = new VarsList(var_names, i18n_names, this.getName() + '_VARS');
        this.setVarsList(va);
        this.fixedPoint_ = PointMass.makeCircle(0.5, 'fixed_point');
        this.fixedPoint_.setMass(Infinity);
        this.bob1_ = PointMass.makeCircle(0.25, 'bob1');
        this.bob1_.setMass(1.0);
        this.bob2_ = PointMass.makeCircle(0.25, 'bob2');
        this.bob2_.setMass(1.0);
        this.spring_ = new Spring('spring', this.fixedPoint_, Vector.ORIGIN, this.bob1_, Vector.ORIGIN, 1.0, 20.0);
        this.getSimList().add(this.fixedPoint_, this.bob1_, this.bob2_, this.stick_, this.spring_);
        va.setValue(THETA, Math.PI * 30 / 180);
        va.setValue(THETAP, 0);
        va.setValue(R, 2.0);
        va.setValue(RP, 0);
        va.setValue(PHI, -Math.PI * 30 / 180);
        va.setValue(PHIP, 0);
        va.setValue(TIME, 0);
        this.saveInitialState();
        this.addParameter(new ParameterNumber(this, DangleStickSim.en.GRAVITY, DangleStickSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(new ParameterNumber(this, DangleStickSim.en.MASS1, DangleStickSim.i18n.MASS1, () => this.getMass1(), a => this.setMass1(a)));
        this.addParameter(new ParameterNumber(this, DangleStickSim.en.MASS2, DangleStickSim.i18n.MASS2, () => this.getMass2(), a => this.setMass2(a)));
        this.addParameter(new ParameterNumber(this, DangleStickSim.en.SPRING_REST_LENGTH, DangleStickSim.i18n.SPRING_REST_LENGTH, () => this.getSpringRestLength(), a => this.setSpringRestLength(a)));
        this.addParameter(new ParameterNumber(this, DangleStickSim.en.SPRING_STIFFNESS, DangleStickSim.i18n.SPRING_STIFFNESS, () => this.getSpringStiffness(), a => this.setSpringStiffness(a)));
        this.addParameter(new ParameterNumber(this, DangleStickSim.en.STICK_LENGTH, DangleStickSim.i18n.STICK_LENGTH, () => this.getStickLength(), a => this.setStickLength(a)));
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + ', gravity_: ' + Util.NF(this.gravity_)
            + ', stickLength_: ' + Util.NF(this.stickLength_)
            + ', spring_: ' + this.spring_
            + ', stick_: ' + this.stick_
            + ', bob1_: ' + this.bob1_
            + ', bob2_: ' + this.bob2_
            + super.toString();
    }
    ;
    getClassName() {
        return 'DangleStickSim';
    }
    ;
    restState() {
        const va = this.getVarsList();
        const vars = va.getValues();
        vars[THETA] = vars[THETAP] = vars[RP] = vars[PHI] = vars[PHIP] = 0;
        const r = this.gravity_ * (this.bob1_.getMass() + this.bob2_.getMass()) /
            this.spring_.getStiffness();
        vars[R] = this.spring_.getRestLength() + r;
        va.setValues(vars);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        const theta_ = Util.limitAngle(vars[THETA]);
        if (theta_ != vars[THETA]) {
            va.setValue(0, theta_, false);
            vars[THETA] = theta_;
        }
        const phi_ = Util.limitAngle(vars[PHI]);
        if (phi_ != vars[PHI]) {
            va.setValue(4, phi_, false);
            vars[PHI] = phi_;
        }
        this.moveObjects(vars);
    }
    ;
    moveObjects(vars) {
        const theta_ = vars[THETA];
        const phi_ = vars[PHI];
        const cosTheta = Math.cos(theta_);
        const sinTheta = Math.sin(theta_);
        const cosPhi = Math.cos(phi_);
        const sinPhi = Math.sin(phi_);
        this.bob1_.setPosition(new Vector(vars[R] * sinTheta, -vars[R] * cosTheta));
        this.bob1_.setVelocity(new Vector(vars[R] * vars[THETAP] * cosTheta + vars[RP] * sinTheta, vars[R] * vars[THETAP] * sinTheta - vars[RP] * cosTheta));
        const L = this.stickLength_;
        this.bob2_.setPosition(new Vector(this.bob1_.getPosition().getX() + L * sinPhi, this.bob1_.getPosition().getY() - L * cosPhi));
        this.bob2_.setVelocity(this.bob1_.getVelocity()
            .add(new Vector(L * vars[PHIP] * cosPhi, L * vars[PHIP] * sinPhi, 0)));
        this.stick_.setStartPoint(this.bob1_.getPosition());
        this.stick_.setEndPoint(this.bob2_.getPosition());
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject == this.bob1_ || simObject == this.bob2_) {
            this.isDragging_ = true;
            return true;
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        const va = this.getVarsList();
        const vars = va.getValues();
        const p = location.subtract(offset);
        if (simObject == this.bob1_) {
            const th = Math.atan2(p.getX(), -p.getY());
            vars[THETA] = th;
            vars[R] = p.length();
            vars[THETAP] = 0;
            vars[RP] = 0;
            vars[PHIP] = 0;
            va.setValues(vars);
        }
        else if (simObject == this.bob2_) {
            const x1 = vars[R] * Math.sin(vars[THETA]);
            const y1 = -vars[R] * Math.cos(vars[THETA]);
            const phi = Math.atan2(p.getX() - x1, -(p.getY() - y1));
            vars[PHI] = phi;
            vars[THETAP] = 0;
            vars[RP] = 0;
            vars[PHIP] = 0;
            va.setValues(vars);
        }
        this.moveObjects(vars);
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.isDragging_ = false;
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        change[TIME] = 1;
        if (!this.isDragging_) {
            const m2 = this.bob2_.getMass();
            const m1 = this.bob1_.getMass();
            const L = this.stickLength_;
            const k = this.spring_.getStiffness();
            const b = this.spring_.getRestLength();
            change[THETA] = vars[THETAP];
            change[R] = vars[RP];
            change[PHI] = vars[PHIP];
            let sum = -4 * m1 * (m1 + m2) * vars[RP] * vars[THETAP];
            sum += 2 * m1 * m2 * L * vars[PHIP] * vars[PHIP] * Math.sin(vars[PHI] - vars[THETA]);
            sum -= 2 * this.gravity_ * m1 * (m1 + m2) * Math.sin(vars[THETA]);
            sum += k * m2 * (b - vars[R]) * Math.sin(2 * (vars[THETA] - vars[PHI]));
            sum = sum / (2 * m1 * (m1 + m2) * vars[R]);
            change[THETAP] = sum;
            sum = 2 * b * k * m1 + b * k * m2 - 2 * k * m1 * vars[R] - k * m2 * vars[R];
            sum -= k * m2 * (b - vars[R]) * Math.cos(2 * (vars[THETA] - vars[PHI]));
            sum += 2 * L * m1 * m2 * Math.cos(vars[PHI] - vars[THETA]) * vars[PHIP] * vars[PHIP];
            sum = sum / (2 * m1 * (m1 + m2));
            sum += vars[R] * vars[THETAP] * vars[THETAP];
            sum += this.gravity_ * Math.cos(vars[THETA]);
            change[RP] = sum;
            change[PHIP] = k * (b - vars[R]) * Math.sin(vars[PHI] - vars[THETA]) / (L * m1);
        }
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.broadcastParameter(DangleStickSim.en.GRAVITY);
    }
    ;
    getMass1() {
        return this.bob1_.getMass();
    }
    ;
    setMass1(value) {
        this.bob1_.setMass(value);
        this.broadcastParameter(DangleStickSim.en.MASS1);
    }
    ;
    getMass2() {
        return this.bob2_.getMass();
    }
    ;
    setMass2(value) {
        this.bob2_.setMass(value);
        this.broadcastParameter(DangleStickSim.en.MASS2);
    }
    ;
    getSpringRestLength() {
        return this.spring_.getRestLength();
    }
    ;
    setSpringRestLength(value) {
        this.spring_.setRestLength(value);
        this.broadcastParameter(DangleStickSim.en.SPRING_REST_LENGTH);
    }
    ;
    getSpringStiffness() {
        return this.spring_.getStiffness();
    }
    ;
    setSpringStiffness(value) {
        this.spring_.setStiffness(value);
        this.broadcastParameter(DangleStickSim.en.SPRING_STIFFNESS);
    }
    ;
    getStickLength() {
        return this.stickLength_;
    }
    ;
    setStickLength(value) {
        this.stickLength_ = value;
        this.broadcastParameter(DangleStickSim.en.STICK_LENGTH);
    }
    ;
}
DangleStickSim.en = {
    SPRING_ANGLE: 'spring angle',
    SPRING_ANGULAR_VELOCITY: 'spring angular velocity',
    SPRING_LENGTH: 'spring length',
    SPRING_LENGTH_VELOCITY: 'spring length velocity',
    STICK_ANGLE: 'stick angle',
    STICK_ANGULAR_VELOCITY: 'stick angular velocity',
    GRAVITY: 'gravity',
    MASS1: 'mass-1',
    MASS2: 'mass-2',
    SPRING_REST_LENGTH: 'spring rest length',
    SPRING_STIFFNESS: 'spring stiffness',
    STICK_LENGTH: 'stick length'
};
DangleStickSim.de_strings = {
    SPRING_ANGLE: 'Federwinkel',
    SPRING_ANGULAR_VELOCITY: 'Federwinkelgeschwindigkeit',
    SPRING_LENGTH: 'Federlänge',
    SPRING_LENGTH_VELOCITY: 'Federlängegeschwindigkeit',
    STICK_ANGLE: 'Stangenwinkel',
    STICK_ANGULAR_VELOCITY: 'Stangenwinkelgeschwindigkeit',
    GRAVITY: 'Gravitation',
    MASS1: 'Masse-1',
    MASS2: 'Masse-2',
    SPRING_REST_LENGTH: 'Federlänge im Ruhezustand',
    SPRING_STIFFNESS: 'Federsteifheit',
    STICK_LENGTH: 'Stangenlänge'
};
DangleStickSim.i18n = Util.LOCALE === 'de' ? DangleStickSim.de_strings : DangleStickSim.en;
Util.defineGlobal('sims$springs$DangleStickSim', DangleStickSim);
