import { AbstractODESim } from '../../lab/model/ODESim.js';
import { EnergyInfo } from '../../lab/model/EnergySystem.js';
import { MoleculeCollision } from './MoleculeCollision.js';
import { MutableVector } from '../../lab/util/MutableVector.js';
import { ParameterNumber } from '../../lab/util/Observe.js';
import { PointMass } from '../../lab/model/PointMass.js';
import { Util } from '../../lab/util/Util.js';
import { VarsList } from '../../lab/model/VarsList.js';
import { Vector } from '../../lab/util/Vector.js';
const TIME = 0;
export const KE = 1;
export const PE = 2;
export const TE = 3;
export const U1X = 4;
export const U1Y = 5;
export const V1X = 6;
export const V1Y = 7;
export class MoleculeSim extends AbstractODESim {
    constructor(opt_name) {
        super(opt_name);
        this.dragAtom_ = -1;
        this.gravity_ = 0;
        this.elasticity_ = 1.0;
        this.damping_ = 0;
        this.distTol_ = 0.02;
        this.timeStep_ = 0.03;
        this.potentialOffset_ = 0;
        this.debugPaint_ = null;
        this.atoms_ = [];
        this.springs_ = [];
        this.collisionFunction_ = null;
        const var_names = [
            VarsList.en.TIME,
            EnergyInfo.en.KINETIC_ENERGY,
            EnergyInfo.en.POTENTIAL_ENERGY,
            EnergyInfo.en.TOTAL_ENERGY,
        ];
        const i18n_names = [
            VarsList.i18n.TIME,
            EnergyInfo.i18n.KINETIC_ENERGY,
            EnergyInfo.i18n.POTENTIAL_ENERGY,
            EnergyInfo.i18n.TOTAL_ENERGY,
        ];
        this.getVarsList().addVariables(var_names, i18n_names);
        this.getVarsList().setComputed(KE, PE, TE);
        this.walls_ = PointMass.makeSquare(12, 'walls');
        this.walls_.setMass(Infinity);
        this.getSimList().add(this.walls_);
        let pn;
        this.addParameter(new ParameterNumber(this, MoleculeSim.en.GRAVITY, MoleculeSim.i18n.GRAVITY, () => this.getGravity(), a => this.setGravity(a)));
        this.addParameter(pn = new ParameterNumber(this, MoleculeSim.en.DAMPING, MoleculeSim.i18n.DAMPING, () => this.getDamping(), a => this.setDamping(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        this.addParameter(pn = new ParameterNumber(this, MoleculeSim.en.ELASTICITY, MoleculeSim.i18n.ELASTICITY, () => this.getElasticity(), a => this.setElasticity(a)));
        pn.setSignifDigits(3);
        pn.setUpperLimit(1);
        this.addParameter(pn = new ParameterNumber(this, EnergyInfo.en.PE_OFFSET, EnergyInfo.i18n.PE_OFFSET, () => this.getPEOffset(), a => this.setPEOffset(a)));
        pn.setLowerLimit(Number.NEGATIVE_INFINITY);
        pn.setSignifDigits(5);
    }
    ;
    toString() {
        return this.toStringShort().slice(0, -1)
            + 'gravity_: ' + Util.NF(this.gravity_)
            + ', damping: ' + Util.NF(this.getDamping())
            + ', elasticity_: ' + Util.NF(this.elasticity_)
            + ', number_of_atoms: ' + this.atoms_.length
            + ', walls_: ' + this.walls_
            + ', potentialOffset_: ' + Util.NF(this.potentialOffset_)
            + super.toString();
    }
    ;
    getClassName() {
        return 'MoleculeSim';
    }
    ;
    addAtom(atom) {
        if (!this.atoms_.includes(atom)) {
            const names = [];
            for (let k = 0; k < 4; k++) {
                names.push(this.getVarName(atom, k, false));
            }
            const localNames = [];
            for (let k = 0; k < 4; k++) {
                localNames.push(this.getVarName(atom, k, true));
            }
            const idx = this.getVarsList().addVariables(names, localNames);
            this.atoms_.push(atom);
            this.getSimList().add(atom);
        }
        this.initializeFromAtom(atom);
    }
    ;
    addSpring(spring) {
        this.springs_.push(spring);
        this.getSimList().add(spring);
    }
    ;
    getAtoms() {
        return Array.from(this.atoms_);
    }
    ;
    getSprings() {
        return Array.from(this.springs_);
    }
    ;
    getVarName(atom, index, localized) {
        let s = atom.getName(localized) + ' ';
        switch (index) {
            case 0:
                s += 'X';
                break;
            case 1:
                s += 'Y';
                break;
            case 2:
                s += 'VX';
                break;
            case 3:
                s += 'VY';
                break;
            default:
                throw '';
        }
        return s;
    }
    ;
    getWalls() {
        return this.walls_;
    }
    ;
    initializeFromAtom(atom) {
        let idx = this.atoms_.indexOf(atom);
        if (idx < 0) {
            throw "atom not found: " + atom;
        }
        idx = 4 * idx;
        const va = this.getVarsList();
        va.setValue(U1X + idx, atom.getPosition().getX());
        va.setValue(U1Y + idx, atom.getPosition().getY());
        va.setValue(V1X + idx, atom.getVelocity().getX());
        va.setValue(V1Y + idx, atom.getVelocity().getY());
        this.getVarsList().incrSequence(KE, PE, TE);
    }
    ;
    setWalls(walls) {
        this.getSimList().remove(this.walls_);
        this.walls_ = walls;
        this.getSimList().add(this.walls_);
    }
    ;
    cleanSlate() {
        const nv = this.getVarsList().numVariables();
        this.getVarsList().setValue(TIME, 0);
        if (nv > U1X) {
            this.getVarsList().deleteVariables(U1X, nv - U1X);
        }
        this.getSimList().removeAll(this.atoms_);
        this.atoms_.length = 0;
        this.getSimList().removeAll(this.springs_);
        this.springs_.length = 0;
    }
    ;
    getEnergyInfo() {
        let ke = 0;
        let pe = 0;
        this.springs_.forEach(spr => pe += spr.getPotentialEnergy());
        const bottom = this.walls_.getBoundsWorld().getBottom();
        this.atoms_.forEach(atom => {
            ke += atom.getKineticEnergy();
            pe += this.gravity_ * atom.getMass() *
                (atom.getPosition().getY() - (bottom + atom.getHeight() / 2));
        });
        return new EnergyInfo(pe + this.potentialOffset_, ke);
    }
    ;
    getPEOffset() {
        return this.potentialOffset_;
    }
    setPEOffset(value) {
        this.potentialOffset_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(EnergyInfo.en.PE_OFFSET);
    }
    ;
    modifyObjects() {
        const va = this.getVarsList();
        const vars = va.getValues();
        this.moveObjects(vars);
        const ei = this.getEnergyInfo();
        va.setValue(KE, ei.getTranslational(), true);
        va.setValue(PE, ei.getPotential(), true);
        va.setValue(TE, ei.getTotalEnergy(), true);
    }
    ;
    moveObjects(vars) {
        this.atoms_.forEach((atom, i) => {
            const idx = 4 * i;
            atom.setPosition(new Vector(vars[U1X + idx], vars[U1Y + idx]));
            atom.setVelocity(new Vector(vars[V1X + idx], vars[V1Y + idx]));
        });
        if (this.debugPaint_ != null) {
            this.debugPaint_();
        }
    }
    ;
    setDebugPaint(fn) {
        this.debugPaint_ = fn;
    }
    ;
    startDrag(simObject, _location, _offset, _dragBody, _modifiers) {
        if (simObject instanceof PointMass) {
            this.dragAtom_ = this.atoms_.indexOf(simObject);
            return this.dragAtom_ > -1;
        }
        else {
            return false;
        }
    }
    ;
    mouseDrag(simObject, location, offset) {
        if (this.dragAtom_ > -1) {
            const atom = this.atoms_[this.dragAtom_];
            if (simObject != atom) {
                return;
            }
            const p = location.subtract(offset);
            let x = p.getX();
            let y = p.getY();
            const w = atom.getWidth() / 2;
            const h = atom.getHeight() / 2;
            const walls = this.walls_.getBoundsWorld();
            if (x < walls.getLeft() + w) {
                x = walls.getLeft() + w + 0.0001;
            }
            if (x > walls.getRight() - w) {
                x = walls.getRight() - w - 0.0001;
            }
            if (y < walls.getBottom() + h) {
                y = walls.getBottom() + h + 0.0001;
            }
            if (y > walls.getTop() - h) {
                y = walls.getTop() - h - 0.0001;
            }
            const va = this.getVarsList();
            const idx = 4 * this.dragAtom_;
            va.setValue(U1X + idx, x);
            va.setValue(U1Y + idx, y);
            va.setValue(V1X + idx, 0);
            va.setValue(V1Y + idx, 0);
            va.incrSequence(KE, PE, TE);
            this.moveObjects(va.getValues());
        }
    }
    ;
    finishDrag(_simObject, _location, _offset) {
        this.dragAtom_ = -1;
        if (!Util.veryDifferent(this.getTime(), 0)) {
            this.saveInitialState();
        }
    }
    ;
    handleKeyEvent(_evt, _pressed, _modifiers) {
    }
    ;
    addCollision(collisions, atom, side, time) {
        const c = new MoleculeCollision(atom, this.walls_, side, time);
        collisions.push(c);
    }
    ;
    findCollisions(collisions, vars, stepSize) {
        this.moveObjects(vars);
        const w = this.walls_.getBoundsWorld();
        this.atoms_.forEach(atom => {
            const a = atom.getBoundsWorld();
            const t = this.getTime() + stepSize;
            if (a.getLeft() < w.getLeft()) {
                this.addCollision(collisions, atom, "left", t);
            }
            if (a.getRight() > w.getRight()) {
                this.addCollision(collisions, atom, "right", t);
            }
            if (a.getBottom() < w.getBottom()) {
                this.addCollision(collisions, atom, "bottom", t);
            }
            if (a.getTop() > w.getTop()) {
                this.addCollision(collisions, atom, "top", t);
            }
        });
    }
    ;
    handleCollisions(collisions, opt_totals) {
        const va = this.getVarsList();
        const vars = va.getValues();
        collisions.forEach(c => {
            const idx = 4 * this.atoms_.indexOf(c.atom);
            switch (c.side) {
                case "left":
                case "right":
                    c.impulse = c.atom.getMass() * (-1 - this.elasticity_) * vars[V1X + idx];
                    va.setValue(V1X + idx, -this.elasticity_ * vars[V1X + idx]);
                    break;
                case "top":
                case "bottom":
                    c.impulse = c.atom.getMass() * (-1 - this.elasticity_) * vars[V1Y + idx];
                    va.setValue(V1Y + idx, -this.elasticity_ * vars[V1Y + idx]);
                    break;
                default:
                    throw '';
            }
            if (opt_totals) {
                opt_totals.addImpulses(1);
            }
            if (this.collisionFunction_ && this.terminal_) {
                this.collisionFunction_(c, this.terminal_);
            }
        });
        va.incrSequence(KE, PE, TE);
        return true;
    }
    ;
    evaluate(vars, change, _timeStep) {
        Util.zeroArray(change);
        this.moveObjects(vars);
        change[TIME] = 1;
        const walls = this.walls_.getBoundsWorld();
        this.atoms_.forEach((atom, listIdx) => {
            if (this.dragAtom_ == listIdx) {
                return;
            }
            const idx = 4 * listIdx;
            const vx = vars[V1X + idx];
            const vy = vars[V1Y + idx];
            change[U1X + idx] = vx;
            change[U1Y + idx] = vy;
            const mass = atom.getMass();
            const bounds = atom.getBoundsWorld();
            const force = new MutableVector(0, 0);
            this.springs_.forEach(spr => {
                if (spr.getBody1() == atom) {
                    force.add(spr.calculateForces()[0].getVector());
                }
                else if (spr.getBody2() == atom) {
                    force.add(spr.calculateForces()[1].getVector());
                }
            });
            force.add(new Vector(0, -this.gravity_ * mass));
            const d = new Vector(vx, vy);
            force.add(d.multiply(-this.damping_));
            let ax = force.getX() / mass;
            if (ax < 0 && Math.abs(bounds.getLeft() - walls.getLeft()) < this.distTol_
                && Math.abs(vx) < -ax * this.timeStep_ / (2 * mass)) {
                ax = 0;
            }
            else if (ax > 0 && Math.abs(bounds.getRight() - walls.getRight()) < this.distTol_
                && Math.abs(vx) < ax * this.timeStep_ / (2 * mass)) {
                ax = 0;
            }
            change[V1X + idx] = ax;
            let ay = force.getY() / mass;
            if (ay < 0 && Math.abs(bounds.getBottom() - walls.getBottom()) < this.distTol_
                && Math.abs(vy) < -ay * this.timeStep_ / (2 * mass)) {
                ay = 0;
            }
            else if (ay > 0 && Math.abs(bounds.getTop() - walls.getTop()) < this.distTol_
                && Math.abs(vy) < ay * this.timeStep_ / (2 * mass)) {
                ay = 0;
            }
            change[V1Y + idx] = ay;
        });
        return null;
    }
    ;
    getGravity() {
        return this.gravity_;
    }
    ;
    setGravity(value) {
        this.gravity_ = value;
        this.getVarsList().incrSequence(PE, TE);
        this.broadcastParameter(MoleculeSim.en.GRAVITY);
    }
    ;
    getDamping() {
        return this.damping_;
    }
    ;
    setDamping(value) {
        this.damping_ = value;
        this.broadcastParameter(MoleculeSim.en.DAMPING);
    }
    ;
    getElasticity() {
        return this.elasticity_;
    }
    ;
    setElasticity(value) {
        this.elasticity_ = value;
        this.broadcastParameter(MoleculeSim.en.ELASTICITY);
    }
    ;
    setCollisionFunction(f) {
        this.collisionFunction_ = f;
    }
    ;
}
MoleculeSim.en = {
    DAMPING: 'damping',
    ELASTICITY: 'elasticity',
    GRAVITY: 'gravity',
};
MoleculeSim.de_strings = {
    DAMPING: 'Dämpfung',
    ELASTICITY: 'Elastizität',
    GRAVITY: 'Gravitation',
};
MoleculeSim.i18n = Util.LOCALE === 'de' ? MoleculeSim.de_strings : MoleculeSim.en;
Util.defineGlobal('sims$springs$MoleculeSim', MoleculeSim);
